import argparse
import sys
import os
from collections import Counter
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

import ncOrtho.coreset.coreset_utils as cu

def main():
    # Print header
    print('\n' + '#' * 38, flush=True)
    print('###' + ' ' * 32 + '###', flush=True)
    print('###   ncOrtho - core set synteny   ###', flush=True)
    print('###' + ' ' * 32 + '###', flush=True)
    print('#' * 38 + '\n', flush=True)

    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description=(
            'Build Covariance models of reference miRNAs from core set of orthologs.'
        )
    )
    parser._action_groups.pop()
    required = parser.add_argument_group('Required Arguments')
    optional = parser.add_argument_group('Optional Arguments')
    # input file path
    required.add_argument('-p', '--parameters', metavar='<path>', type=str, required=True,
                          help='Path to the parameters file in yaml format')
    optional.add_argument(
        '--max_anchor_dist', metavar='int', type=int,
        help='Number of additional genes to the left and right '
             'of the reference miRNA that are to be considered as syntenic anchors. (Default: 3)',
        nargs='?', const=3, default=3
    )
    optional.add_argument(
        '--idtype', metavar='str', type=str,
        help='Choose the ID in the reference gff file that is '
             'compared to the IDs in the pairwise orthologs file (default:GeneID) '
             'Options: ID, Name, GeneID, gene_id',
        nargs='?', const='ID=', default='ID'
    )
    # Show help when no arguments are added.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()

    # parameters
    add_pos_orthos = args.max_anchor_dist
    idtype = args.idtype

    core_dict, ref_paths, all_paths = cu.parse_yaml(args.parameters)
    # check if files exist
    for cp in all_paths:
        if not os.path.isfile(cp):
            raise ValueError(f'{cp} does not exist')

    print('# Reading pairwise orthologs', flush=True)
    ortho_dict = cu.read_pairwise_orthologs(core_dict)


    def neighbors(prot, anno, anchordist):
        if prot not in anno:  # pseudogenes are not in parsed annotation file but can occur in ortholog file
            return None
        chrom, position = anno[prot]
        anchors = []
        for index in range(position - anchordist, position + anchordist + 1):
            if index not in anno[chrom]:  # if gene positioned at end of contig, no more anchors can be considered
                continue
            anchor, a_start, a_end, a_strand = anno[chrom][index]
            anchors.append(anchor)
        anchors.remove(prot)
        return anchors


    def synteny_distance(block, ref_prot, coreanno, orthodict, anchordist, corespecies, mgi=3):
        """
        returns: Number of conserved anchors per number of anchors
        """
        o = []

        # find position of reference protein orthologs in core species
        reforthos = orthodict[ref_prot]
        for refortho in reforthos:
            conserved_anchors = 0

            if refortho not in coreanno:  # pseudogenes are not in parsed annotation file but can occur in ortholog file
                continue
            chrom, position = coreanno[refortho]

            # find position of anchor protein orthologs in core species
            for anchor in block:

                # find orthologs of anchor proteins
                if anchor not in orthodict:  # synteny not fulfilled if no ortholog in core species
                    # c.append(0)
                    continue
                anchor_ortho_list = orthodict[anchor]
                for anchor_ortho in anchor_ortho_list:
                    if anchor_ortho not in coreanno:  # e.g. for pseudogenes
                        # c.append(0)
                        continue

                    a_chrom, a_position = coreanno[anchor_ortho]
                    # if distance to reference protein is similar in core species than in reference species, accept
                    if a_chrom == chrom and abs(int(position) - int(a_position)) <= anchordist + mgi:
                        conserved_anchors += 1
                        break

                    # c.append(0)
            # o.append(sum(c) / len(c))
            o.append([corespecies, refortho, conserved_anchors])
        return o


    print('# Evaluating synteny')
    ref_anno_dict = cu.parse_annotation(ref_paths['annotation'], idtype)
    #syn_col = {}
    syn_col = []
    for corespec, pairwiseorthos in ortho_dict.items():
        #core_col = []
        print(corespec)

        core_anno_dict = cu.parse_annotation(core_dict[corespec]['annotation'], idtype)

        for refprot in pairwiseorthos:
            syn_block = neighbors(refprot, ref_anno_dict, add_pos_orthos)
            if not syn_block:  # in rare case that reference protein from orthology file is not found in annotation
                continue
            syn_fullfilled_list = synteny_distance(syn_block, refprot, core_anno_dict, pairwiseorthos, add_pos_orthos, corespec)
            syn_col.extend(syn_fullfilled_list)

        #syn_col[corespec] = dict(Counter(core_col))

    #print(syn_col)
    df = pd.DataFrame(syn_col, columns=['species', 'reference_protein', 'num_conserved_anchors'])
    #display(df)

    o = ['Gorilla_gorilla', 'Macaca_mulatta', 'Pongo_abelii', 'Nomascus_leucogenys', 'Saimiri_boliviensis',
         'Mus_musculus', 'Cavia_porcellus', 'Canis_familiaris', 'Gallus_gallus']
    sns.set(rc={'figure.figsize': (12, 4), 'ytick.left': True, 'xtick.bottom': True}, font_scale=1.2, style='whitegrid')
    ax = sns.histplot(data=df, x='num_conserved_anchors', hue='species', discrete=True, multiple='dodge', hue_order=o,
                      shrink=0.8)
    sns.move_legend(ax, "upper left", bbox_to_anchor=(1, 1))
    plt.show()



if __name__ == '__main__':
    main()


import glob
import subprocess as sp


def check_blastdb(db_path):
    file_extensions = ['nhr', 'nin', 'nsq']
    for fe in file_extensions:
        files = glob.glob(f'{db_path}*{fe}')
        if not files:
            # At least one of the BLAST db files is not existent
            return False
    return True


def make_blastndb(inpath, outpath):
    db_command = 'makeblastdb -in {} -out {} -dbtype nucl'.format(inpath, outpath)
    sp.call(db_command, shell=True)


def find_refbit(outstr):
    for line in outstr.split('\n'):
        if line.startswith('  '):
            highest_score = float(line.split()[3])
            return highest_score
    # if no results return topscore 0
    print(
        '# Warning: Self bit score not applicable, '
        'setting threshold to 0'
    )
    highest_score = 0.0
    return highest_score


def vprint(s, verbose):
    if verbose:
        print(s, flush=True)



"""
ncOrtho - Targeted ortholog search for miRNAs
Copyright (C) 2021 Felix Langschied

ncOrtho is a free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ncOrtho is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ncOrtho.  If not, see <http://www.gnu.org/licenses/>.
"""

# Create a core set of orthologs
# Find the corresponding syntenic regions in reference and core species
# Search for core orthologs by reciprocal BLAST search
# Create Stockholm structural alignment

# required:
# reference microRNA data (sequence, coordinates)
# reference taxon: genome, blastdb, gtf file with gene coordinates
# core set taxa: genome, gtf file, pairwise orthologs

import argparse
import multiprocessing as mp
import os
import sys
import shutil

try:
    from createcm import CmConstructor
    from core_reblast import blastsearch
    from locate_position import categorize_mirna_position
    from synteny import analyze_synteny
    import coreset_utils as cu
except ModuleNotFoundError:
    from ncOrtho.coreset.createcm import CmConstructor
    from ncOrtho.coreset.core_reblast import blastsearch
    from ncOrtho.coreset.synteny import analyze_synteny
    from ncOrtho.coreset.locate_position import categorize_mirna_position
    import ncOrtho.coreset.coreset_utils as cu


###############################################################################
def main():
    # Print header
    print('\n' + '#' * 43, flush=True)
    print('###' + ' ' * 37 + '###', flush=True)
    print('###   ncOrtho - core set construction   ###', flush=True)
    print('###' + ' ' * 37 + '###', flush=True)
    print('#' * 43 + '\n', flush=True)

    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description=(
            'Build Covariance models of reference miRNAs from core set of orthologs.'
        )
    )
    parser._action_groups.pop()
    required = parser.add_argument_group('Required Arguments')
    optional = parser.add_argument_group('Optional Arguments')
    # input file path
    required.add_argument('-p', '--parameters', metavar='<path>', type=str, required=True,
                          help='Path to the parameters file in yaml format')
    # mirna data
    required.add_argument(
        '-n', '--ncrna', metavar='<path>', type=str, required=True,
        help='Path to tab separated file of reference miRNAs information '
    )
    # output folder
    required.add_argument(
        '-o', '--output', metavar='<path>', type=str, required=True,
        help='Path for the output folder'
    )
    # OPTIONAL VARIABLES
    # cpu, use maximum number of available cpus if not specified otherwise
    optional.add_argument(
        '--threads', metavar='int', type=int,
        help='Number of CPU cores to use (Default: All available)', nargs='?',
        const=mp.cpu_count(), default=mp.cpu_count()
    )
    # Maximum gene insertions
    optional.add_argument(
        '--mgi', metavar='int', type=int,
        help='Maximum number of gene insertions in the core species (Default: 3)', nargs='?',
        const=3, default=3
    )
    # use dust filter?
    optional.add_argument(
        '--dust', metavar='yes/no', type=str,
        help='Use BLASTn dust filter. '
             'Will not build models from repeat regions if "yes". (Default: no)',
        nargs='?',
        const='no', default='no'
    )
    optional.add_argument(
        '--create_model', metavar='yes/no', type=str,
        help='set to "no" if you only want to create the alignment. (Default: yes)', nargs='?',
        const='yes', default='yes'
    )
    #
    optional.add_argument(
        '--idtype', metavar='str', type=str,
        help='Choose the ID in the reference gff file that is '
             'compared to the IDs in the pairwise orthologs file (default:GeneID) '
             'Options: ID, Name, GeneID, gene_id, CDS',
        nargs='?', const='ID=', default='ID'
    )
    optional.add_argument(
        '--max_anchor_dist', metavar='int', type=int,
        help='Number of additional genes to the left and right '
             'of the reference miRNA that are to be considered as syntenic anchors. (Default: 3)',
        nargs='?', const=3, default=3
    )
    optional.add_argument(
        '--verbose', metavar='yes/no', type=str,
        help='Print additional information (Default: no)',
        nargs='?', const='no', default='no'
    )

    ###############################################################################

    # Show help when no arguments are added.
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()

    # Check if computer provides the desired number of cores.
    available_cpu = mp.cpu_count()
    if args.threads > available_cpu:
        raise ValueError('The provided number of CPU cores is higher than the number available on this system')
    else:
        cpu = args.threads

    # parse mandatory arguments
    output = os.path.realpath(args.output)

    # parse optional arguments
    # mgi = args.mgi
    dust = args.dust
    create_model = args.create_model
    if args.verbose == 'yes':
        verbose = True
    elif args.verbose == 'no':
        verbose = False
    else:
        raise ValueError(f'Unknown value for "verbose": {args.verbose}')
    add_pos_orthos = args.max_anchor_dist

    # parameters
    core_dict, ref_paths, all_paths = cu.parse_yaml(args.parameters)
    # check if files exist
    for cp in all_paths:
        if not os.path.isfile(cp):
            raise ValueError(f'{cp} does not exist')

    ###############################################################################
    neighbor_dict = {}
    mirna_dict = {}

    # create directory for intermediate files
    tmpout = f'{output}/tmp'
    if not os.path.isdir(tmpout):
        os.makedirs(tmpout)

    print('# Reading pairwise orthologs', flush=True)
    if args.idtype == 'CDS':
        ortho_dict = cu.pairwise_orthologs_from_cds(core_dict, ref_paths['annotation'])
    else:
        ortho_dict = cu.read_pairwise_orthologs(core_dict)

    print('# Reading miRNA data', flush=True)
    mirnas = cu.read_mirnas(args.ncrna)

    print('# Reading reference annotation', flush=True)
    ref_dict = cu.parse_annotation(ref_paths['annotation'], args.idtype)

    print('# Determining the position of each miRNA and its neighboring gene(s)', flush=True)
    mirna_positions = {}
    for mirna in mirnas:
        mirid, chromo, start, end, strand = cu.mirna_position(mirna)
        syntype, core_orthos = categorize_mirna_position(
            mirid, chromo, start, end, strand, ref_dict, ortho_dict, add_pos_orthos, verbose
        )
        if not syntype:
            print(f'Warning: Could not localize {mirid}')
            continue
        mirna_positions[mirid] = (syntype, core_orthos)

    print('### Identifying syntenic regions in core species', flush=True)
    syntenyregion_per_mirna = analyze_synteny(core_dict, mirna_positions, tmpout, args.idtype, args.mgi, verbose)

    if not syntenyregion_per_mirna:
        raise ValueError('No regions of conserved synteny found in any core species')

    for mirid, fastalist in syntenyregion_per_mirna.items():
        if not fastalist:
            print(f'Warning: No syntenic region found in any core species for {mirid}. '
                  f'Make sure that the IDs in the annotation file match the ones in the orthologs file')
            continue
        miroutdir = f'{tmpout}/{mirid}'
        if not os.path.isdir(miroutdir):
            os.mkdir(miroutdir)
        with open(f'{miroutdir}/synteny_regions_{mirid}.fa', 'w') as of:
            for line in fastalist:
                of.write(line)

    #################################################################################################
    print('\n### Starting Ortholog search')
    for mirna in mirnas:
        mirid = mirna[0]
        sto_path = blastsearch(mirna, ref_paths['genome'], tmpout, cpu, dust, verbose)
        if create_model == 'yes' and sto_path is not None:
            print(f'# Starting to construct covariance model for {mirid}', flush=True)
            model_out = f'{output}/{mirid}.cm'
            if not os.path.isfile(model_out):
                # Initiate covariance model construction and calibration.
                cmc = CmConstructor(sto_path, output, mirid, cpu)
                # Construct the model.
                cmc.construct()
                # Calibrate the model.
                cmc.calibrate()
            else:
                print(f'Model of {mirid} already found at {output}. Nothing done..', flush=True)
    print('\n### Construction of core set finished')


if __name__ == '__main__':
    main()



# ### find left neighbor or check if located inside gene
# ### chr_dict[contig][i] = (geneid, start, end, strand)
# ###############################################################################
# # case 1): there is no protein-coding gene on the same contig as the miRNA,
# # so there can be no neighbors (should only occur in highly fragmented
# # assemblies)
# if not chromo in ref_dict.keys():
#     print(
#         'WARNING: No protein-coding genes found on contig "{0}". '
#         'Synteny around {1} cannot be established.\n'
#         'Make sure that the contig identifiers of the miRNA input file '
#         'match the ones in the reference annotation file'
#             .format(chromo, mirid)
#     )
#     continue
#
# # case 2): miRNA is located left of the first gene and hence has no left
# # neighbor, the first gene is therefore by default the right neighbor
# if end < int(ref_dict[chromo][1][1]):
#     print(
#         'There is no left neighbor of {0}, since it is located at the '
#         'start of contig {1}.'.format(mirid, chromo)
#     )
#     print(
#         '{0} is the right neighbor of {1}.'
#             .format(ref_dict[chromo][1][0], mirid)
#     )
#     continue
#
# # case 3): miRNA is located right to the last gene, so the last gene is the
# # left neighbor and there cannot be a right neighbor
# elif start > int(ref_dict[chromo][len(ref_dict[chromo])][2]):
#     print(
#         '{0} is the left neighbor of {1}.'
#             .format(ref_dict[chromo][len(ref_dict[chromo])][0], mirid)
#     )
#     print(
#         'There is no right neighbor of {0}, since it is located at the'
#         ' end of contig {1}.'.format(mirid, chromo)
#     )
#     continue
#
# # case 4): miRNA is located either between two genes or overlapping with (an
# # intron of) a gene, either on the same or the opposite strand
# ###############################################################################
# else:
#     solved = False
#     for i, gene in enumerate(ref_dict[chromo]):
#         gene_data = ref_dict[chromo][gene]
#         # case 4.1): miRNA inside gene
#         if (
#                 start >= gene_data[1]
#                 and end <= gene_data[2]
#                 and strand == gene_data[3]
#         ):
#             solved = True
#             # c+=1
#             print(
#                 '{0} is located inside the gene {1}.'
#                     .format(mirid, gene_data[0])
#             )
#             ortho_hits = ortho_search(gene_data[0], ortho_dict)
#             for core_tax in ortho_hits:
#                 try:
#                     neighbor_dict[core_tax][mirid] = (
#                         ('inside', ortho_hits[core_tax])
#                     )
#                 except:
#                     neighbor_dict[core_tax] = (
#                         {mirid: ('inside', ortho_hits[core_tax])}
#                     )
#             break
#         # case 4.2): miRNA opposite of gene
#         elif (
#                 start >= gene_data[1]
#                 and end <= gene_data[2]
#                 and strand != gene_data[3]
#         ):
#             solved = True
#             print(
#                 '{0} is located opposite of the gene {1}.'
#                     .format(mirid, gene_data[0])
#             )
#             ortho_hits = ortho_search(gene_data[0], ortho_dict)
#             for core_tax in ortho_hits:
#                 try:
#                     neighbor_dict[core_tax][mirid] = (
#                         ('opposite', ortho_hits[core_tax])
#                     )
#                 except:
#                     neighbor_dict[core_tax] = (
#                         {mirid: ('opposite', ortho_hits[core_tax])}
#                     )
#             break
#         # case 4.3): miRNA between genes
#         elif (
#                 int(ref_dict[chromo][gene][2]) < start
#                 and ref_dict[chromo][gene + 1][1] > end
#         ):
#             solved = True
#             ###############################################################################
#             print(
#                 '{1} is the left neighbor of {2}.'
#                 .format(gene, ref_dict[chromo][gene][0], mirid)
#             )
#             print(
#                 '{1} is the right neighbor of {2}.'
#                 .format(gene, ref_dict[chromo][gene + 1][0], mirid)
#             )
#             left_hits = ortho_search(gene_data[0], ortho_dict)
#             right_hits = (
#                 ortho_search(ref_dict[chromo][gene + 1][0], ortho_dict)
#             )
#             # save only the hits where both genes have orthologs in a species
#             if left_hits:
#                 # print(left_hits)
#                 # print(right_hits)
#                 for taxon in left_hits:
#                     if taxon in right_hits:
#                         try:
#                             neighbor_dict[taxon][mirid] = (
#                                 (
#                                     'in-between',
#                                     [left_hits[taxon],
#                                      right_hits[taxon]]
#                                 )
#                             )
#                         except:
#                             neighbor_dict[taxon] = (
#                                 {mirid: (
#                                     'in-between',
#                                     [left_hits[taxon],
#                                      right_hits[taxon]]
#                                 )}
#                             )
#                     else:
#                         print('Orthologs were not found for both flanking genes')
#             break
#     if not solved:
#         print('Unable to resolve synteny for {}.'.format(mirid))
#
#
# ##########################################################################################
#
#         print('# Loading genome file')
#         fasta_path = core_dict[taxon]['genome']
#         core_gen_dict = '{}/core_genomes'.format(output)
#         if not os.path.isdir(core_gen_dict):
#             os.mkdir(core_gen_dict)
#         slink = '{}/slink_to_{}'.format(core_gen_dict, taxon)
#         try:
#             os.symlink(fasta_path, slink)
#         except FileExistsError:
#             pass
#         genome = pyfaidx.Fasta(slink)
#         print('Done')
#
#
# ##########################################################################################
# # print(mirna)
# style = neighbor_dict[taxon][mirna][0]
# if style == 'inside' or style == 'opposite':
#     print(f'# {mirna} location: {style} of gene')
#     try:
#         ortho_data = (
#             core_dict[neighbor_dict[taxon][mirna][1]]
#         )
#     except KeyError as e:
#         print('{} not found in annotation file.'.format(e))
#         continue
#     positions = list(
#         core_dict[ortho_data[0]][ortho_data[1]][1:4]
#     )
#     coordinates = [ortho_data[0]] + positions
#     seq = (
#         genome[coordinates[0]]
#         [coordinates[1] - 1:coordinates[2]].seq
#     )
#     # print(seq[0:10])
#     try:
#         mirna_dict[mirna][taxon] = seq
#     except KeyError:
#         mirna_dict[mirna] = {taxon: seq}
# ##########################################################################################
#
#     # Search for the coordinates of the orthologs and extract the sequences
#     for taxon in neighbor_dict:
#         # print('YOU MADE IT THIS FAR.')
#         for mirna in neighbor_dict[taxon]:
#
#             if style == 'in-between':
#                 print(f'# {mirna} location: between genes')
#                 try:
#                     left_data = (
#                         core_dict[neighbor_dict[taxon][mirna][1][0]]
#                     )
#                     right_data = (
#                         core_dict[neighbor_dict[taxon][mirna][1][1]]
#                     )
#                 except KeyError as e:
#                     print('{} not found in annotation file.'.format(e))
#                     continue
#                 # Test to see if the two orthologs are themselves neighbors where their
#                 # distance cannot be larger than the selected mgi value. This accounts
#                 # for insertions in the core species.
#                 # TODO: Apply mgi also to the reference species to account for insertions
#                 # in the reference.
#                 #                 print(f'left_data: {left_data}')
#                 #                 print(f'right_data: {right_data}')
#                 if (
#                         left_data[0] == right_data[0]
#                         and abs(left_data[1] - right_data[1]) <= mgi
#                 ):
#                     # Determine which sequence to include for the synteny-based ortholog search
#                     # depending on the order of orthologs. The order of the orthologs in the core
#                     # species might be inverted compared to that in the reference species.
#                     ###############################################################################
#                     if left_data[1] < right_data[1]:
#                         # print('left')
#                         # print(core_gtf_dict[left_data[0]][left_data[1]])
#                         # print(core_gtf_dict[right_data[0]][right_data[1]])
#                         contig = left_data[0]
#                         # print(contig)
#                         seq_start = int(
#                             core_dict[left_data[0]][left_data[1]][2]
#                         )
#                         # print(seq_start)
#                         seq_end = (
#                             core_dict[right_data[0]][right_data[1]][1]
#                         )
#                         # print(seq_end)
#                         seq = genome[contig][seq_start - 1:seq_end].seq
#                         try:
#                             mirna_dict[mirna][taxon] = seq
#                         except:
#                             mirna_dict[mirna] = {taxon: seq}
#                     elif right_data[1] < left_data[1]:
#                         # print('right')
#                         # print(core_gtf_dict[left_data[0]][left_data[1]])
#                         # print(core_gtf_dict[right_data[0]][right_data[1]])
#                         contig = left_data[0]
#                         # print(contig)
#                         # seq_start = int(
#                         #     core_gtf_dict[left_data[0]][left_data[1]][2]
#                         # )
#                         # print(seq_start)
#                         # seq_end = (
#                         #     core_gtf_dict[right_data[0]][right_data[1]][1]
#                         # )
#                         seq_start = int(
#                             core_dict[right_data[0]][right_data[1]][2]
#                         )
#                         # print(seq_start)
#                         seq_end = (
#                             core_dict[left_data[0]][left_data[1]][1]
#                         )
#                         # print(seq_end)
#                         seq = genome[contig][seq_start - 1:seq_end].seq
#                         try:
#                             mirna_dict[mirna][taxon] = seq
#                         except:
#                             mirna_dict[mirna] = {taxon: seq}
#                     print('Synteny fulfilled.')
#                 else:
#                     print(
#                         'No shared synteny for {} in {}.'
#                             .format(mirna, taxon)
#                     )
#                     continue
#                     # print(left_data)
#                     # print(right_data)
#             else:
#                 print('## Neither inside, opposite, nor in-between')
#                 # print(neighbor_dict[taxon][mirna])
#                 continue
#             print('Candidate region found')
#
#             ##########################################################################################
#             # Write output file
#             for mirna in mirna_dict:
#                 # print('{0}/{1}/{1}.fa'.format(output, mirna))
#                 with open('{0}/{1}/{1}.fa'.format(output, mirna), 'w') as outfile:
#                     for core_taxon in mirna_dict[mirna]:
#                         outfile.write(
#                             '>{0}\n{1}\n'
#                                 .format(core_taxon, mirna_dict[mirna][core_taxon])
#                         )
#             if not mirna_dict:
#                 print('\nWARNING: No syntenic regions found in the core species')
#             ##########################################################################################
#
#             # skip_file = '{}/not_found_in_ref.fa'.format(o_path)
#             # if not os.path.isfile(skip_file):
#             #     with open(skip_file, 'w') as of:
#             #         of.write('>{}\n{}\n'.format(mirna, seq))
#             # else:
#             #     with open(skip_file, 'a') as of:
#             #         of.write('>{}\n{}\n'.format(mirna, seq))
#             # return None
#
#     ##########################################################################################
#     ##### Collect best hit for each core set species if it is within the accepted bit score range
#
#     core_dict = {}
#     result_list = results.split('\n')
#     # print(result_list)
#     if result_list:
#         #print(result_list)
#         for hit in result_list:
#             if hit:
#                 hit_data = hit.split()
#                 # print(hit_data)
#                 # print(f'hit bitscore: {hit_data[2]}')
#                 # print(f'you needed: {0.5*ref_bit_score}')
#                 # print(f'and you got: {hit_data[2]}')
#                 if (
#                     not hit_data[0] in core_dict
#                     and float(hit_data[2]) >= 0.5*ref_bit_score
#                 ):
#                     core_dict[hit_data[0]] = hit
#                     print(f'pre-miRNA candidate found for {hit_data[0]}! ({hit_data[2]}/{0.5*ref_bit_score})')
#                     break
#                 else:
#                     print(f'Syntenic candidate region BLAST hit below reference bit score threshold ({hit_data[2]}/{0.5*ref_bit_score})')
#
#     if len(core_dict) == 0:
#         print('WARNING: No region in the core species scored above the reference bit score threshold')
#
#         ##### Re-BLAST #####
#         print('\n### Starting reciprocal BLAST search.')
#         accept_dict = {}
#         for species in core_dict.keys():
#             print(f'# {species}')
#             # Make sure to eliminate gaps
#             candidate_seq = core_dict[species].split()[3].replace('-', '')
#             reblastn_cmd = (
#                 'blastn -num_threads {c} -task blastn -dust {dust} -db {ref_blastdb} -outfmt \"6'
#                 ' sseqid sstart send evalue bitscore\"'
#             )
#             reblastn = sp.Popen(
#                 reblastn_cmd, shell=True, stdin=sp.PIPE,
#                 stdout=sp.PIPE, stderr=sp.STDOUT, encoding='utf8'
#             )
#             reresults, reerr = reblastn.communicate(candidate_seq)
#             if reerr:
#                 print(reerr)
#                 sys.exit()
#
#             # Check if reverse hit overlaps with reference miRNA
#             if reresults:
#                 first_hit = reresults.split('\n')[0].split()
#                 rchr = first_hit[0]
#                 rstart = int(first_hit[1])
#                 rend = int(first_hit[2])
#                 if rchr == mchr:
#                     # print('Same chromosome.')
#                     if (
#                             (rstart <= mstart and mstart <= rend)
#                             or (rstart <= mend and mend <= rend)
#                     ):
#                         # first within second
#                         print('Reciprocity fulfilled.')
#                         accept_dict[species] = core_dict[species]
#                     elif (
#                             (mstart <= rstart and rstart <= mend)
#                             or (mstart <= rend and rend <= mend)
#                     ):
#
#                         # second within first
#                         print('Reciprocity fulfilled.')
#                         accept_dict[species] = core_dict[species]
#                     else:
#                         print('Reciprocity unfulfilled.')
#                 else:
#                     print('Hit on chromosome {}, expected {}'.format(rchr, mchr))
#             else:
#                 print(
#                     'No reverse hit for {}. Reciprocity unfulfilled.'
#                         .format(mirid)
#
#
#
#     with open(corefile, 'w') as outfile:
#         outfile.write('>{}\n{}\n'.format(mirid, preseq))
#         print('\n### Starting Alignment')
#         # print('>{}\n{}'.format(mirid, preseq))
#         for accepted in accept_dict:
#             # print('>{}\n{}'.format(accepted, core_dict[accepted].split('\t')[3]))
#             outfile.write(
#                 '>{}\n{}\n'
#                     .format(accepted, accept_dict[accepted].split('\t')[3])
#                     .replace('-', '')


import glob
import os
import tempfile
import subprocess as sp
import sys
import shutil
import inspect
from collections import Counter
import argparse

"""
Species in spec_to_skip will not be part of multiple sequence alignments
"""


def make_supermatrix(out):
    align_out = '{}/alignments'.format(out)
    tree_out = '{}/supermatrix'.format(out)
    # curr_dir = sys.path[0]
    curr_dir = '/'.join(inspect.getfile(inspect.currentframe()).split('/')[:-1])
    if not os.path.isdir(tree_out):
        os.mkdir(tree_out)
    print('# Creating supermatrix alignment')
    cmd = (
        'perl {}/concat_alignments_dmp.pl -in {} -out supermatrix.aln'
            .format(curr_dir, align_out)
    )
    res = sp.run(cmd, shell=True, capture_output=True)
    if res.returncode != 0:
        print('ERROR:')
        print(res.stderr.decode('utf-8'))
        sys.exit()
    # print('# Done')

    print('# De-gapping alignment')
    cmd = (
        'perl {}/degapper.pl -in {}/supermatrix.aln'.format(curr_dir, out)
    )
    sp.run(cmd, shell=True)
    # move files to output dir
    shutil.move(f'{out}/supermatrix.aln', f'{tree_out}/supermatrix.aln')
    shutil.move(f'{out}/subalignment_positions.txt', f'{tree_out}/subalignment_positions.txt')
    shutil.move(f'{out}/supermatrix.aln.proc', f'{tree_out}/supermatrix.aln.proc')
    # print('# Done')
    # return f'{tree_out}/supermatrix.aln'
    return f'{tree_out}/supermatrix.aln.proc'


def arglistcheck(arg):
    if not arg:
        return ''
    if os.path.isfile(arg):
        with open(arg) as fh:
            arglist = [element.strip() for element in fh.read().split('\n') if element]
    else:
        arglist = [element.strip() for element in arg.split(',') if element]
    return arglist


def parse_matseq(path):
    m2s = {}
    with open(path) as fh:
        for line in fh:
            if line.startswith('#'):
                continue
            mature = line.strip().split('\t')[-1].replace('U', 'T')
            mirid = line.strip().split('\t')[1].split('_')[0]
            if len(mature) < 7:
                raise ValueError('No mature sequence found in "ncrna_file"')
            seed = mature[1:8]
            m2s[mirid] = seed
    return m2s

def main():
    # Print header
    print('\n'+'#'*39)
    print('###'+' '*33+'###')
    print('###   Analysis of ncOrtho results   ###')
    print('###'+' '*33+'###')
    print('#'*39+'\n')
    sys.stdout.flush()

    # Parse command-line arguments
    # Define global variables
    parser = argparse.ArgumentParser(
        description='Analzye ncOrtho results. Gives: PhyloProfile input, supermatrix alignment, species tree.'
    )
    parser._action_groups.pop()
    required = parser.add_argument_group('Required Arguments')
    optional = parser.add_argument_group('Optional Arguments')
    required.add_argument(
        '-r', '--results', metavar='<path>', type=str,
        help='Path to ncOrtho output directory that is to be analyzed'
    )
    required.add_argument(
        '-o', '--output', metavar='<path>', type=str,
        help='Path to location where output of analysis should be stored. '
             'Assumes that result files end with "_orthologs.fa"'
    )
    required.add_argument(
        '-m', '--mapping', metavar='<path>', type=str,
        help='Tab seperated mapping file between NCBI taxonomy id and name of the query species '
             'or more generally the name of the directory that contains the ncOrtho results of each species'
             '(e.g "9606\tHomo_sapiens")'
    )
    ##########################################################################################
    optional.add_argument(
        '--skip', metavar='str', type=str, nargs='?', const='', default='',
        help=(
            'Comma separated list or path to newline seperated list of species for which analyses should be skipped'
        )
    )
    optional.add_argument(
        '--include', metavar='str', type=str, nargs='?', const='', default='',
        help=(
            'Comma separated list or path to newline seperated list of species for which analyses should be performed. '
            'Will skip every species not in list.'
        )
    )
    optional.add_argument(
        '--auto_skip', metavar='float', type=float, nargs='?', const=0.0, default=0.0,
        help=(
            'Exclude species for which less than a fraction of the reference '
            'miRNAs were detected from species tree calculation. '
            '(e.g. --auto_skip 0.5 means that only species in which at least 50%% '
            'of the reference miRNAs were found are used for species tree calculation) (Default: Off)'
        )
    )
    optional.add_argument(
        '--iqtree', metavar='str', type=str, nargs='?',
        const='iqtree -s {} -bb 1000 -alrt 1000 -nt AUTO -redo -pre {}/species_tree',
        default='iqtree -s {} -bb 1000 -alrt 1000 -nt AUTO -redo -pre {}/species_tree',
        help=(
            'Call to iqtree. Do not change curly bracket notation '
            '(Default: iqtree -s {} -bb 1000 -alrt 1000 -nt AUTO -redo -pre {}/{})'
        )
    )
    optional.add_argument(
        '--mirnas', metavar='<path>', type=str, nargs='?', const='', default='',
        help=(
            'Comma separated list or path to newline seperated file with miRNAs that should be used for reconstruction'
        )
    )
    # mirna data
    optional.add_argument(
        '--ncrna_file', metavar='<path>', nargs='?', const='', default='',
        help='Path to tab separated file of reference miRNAs information, as used in ncCreate and ncSearch. '
             'Uses mature sequence column to identify seed conservation.'
    )
    ##########################################################################################
    # Parse arguments
    ##########################################################################################
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    else:
        args = parser.parse_args()

    res_dict = args.results
    taxmap = args.mapping
    outdir = args.output

    spec_to_skip = arglistcheck(args.skip)
    spec_include = arglistcheck(args.include)

    overlap = set(spec_to_skip).intersection(set(spec_include))
    if overlap:
        raise ValueError(f'"{overlap}" present in species to skip as well as species to include')


    mirlist = arglistcheck(args.mirnas)

    iqtree_cmd = args.iqtree
    auto_skip = args.auto_skip

    if args.ncrna_file:
        mirid2seed = parse_matseq(args.ncrna_file)
    else:
        mirid2seed = {}
    ##########################################################################################
    # Argument checks
    ##########################################################################################
    if not os.path.isdir(outdir):
        os.mkdir(outdir)
    if auto_skip < 0 or auto_skip > 1:
        raise ValueError(f'--auto_skip must be a value between 0 and 1, not {auto_skip}')

    ###############################################################################################
    # Main
    ###############################################################################################
    # find results
    ortholog_files = glob.glob(f'{res_dict}/*/*_orthologs.fa')
    if not ortholog_files:
        ortholog_files = glob.glob(f'{res_dict}/*.fa')
    # read mapping file
    name_2_taxid = {}
    with open(taxmap, 'r') as mp:
        for line in mp:
            taxid, name = line.strip().split('\t')
            name_2_taxid[name] = taxid

    ortho_dict = {}  # ortho_dict = {mirna: {species: seq}}
    # read all results into one dictionary and write phyloprofile input
    pp_in = f'{outdir}/PhyloProfile.long'
    spec_list = []
    with open(pp_in, 'w') as pp:
        pp.write('geneID\tncbiID\torthoID\n')
        for file in ortholog_files:
            with open(file, 'r') as fh:
                for line in fh:
                    if line.startswith('>'):
                        spec, mirna = line.strip().replace('>', '').split('|')[:2]
                    else:
                        seq = line.strip()
                        # write phyloprofile input
                        taxstr = f'ncbi{name_2_taxid[spec]}'
                        group_str = mirna.split('_')[0]
                        if mirid2seed:
                            seed = mirid2seed[group_str]
                            if seed in seq:
                                seedcon = 1
                            else:
                                seedcon = 0
                            pp.write(f'{group_str}\t{taxstr}\t{mirna}\t{seedcon}\n')
                        else:
                            pp.write(f'{group_str}\t{taxstr}\t{mirna}\n')
                        # initialize dictionary
                        if group_str not in ortho_dict:
                            ortho_dict[group_str] = {}
                        # only save representative (first) hit
                        if spec not in ortho_dict[group_str]:
                            ortho_dict[group_str][spec] = seq
                            spec_list.append(spec)
    print('# Finished writing PhyloProfile input')

    # if miRNAfile option is chosen, prune dictionary
    if mirlist:
        tmpdict = {}
        for mirnaid, value in ortho_dict.items():
            if mirnaid in mirlist:
                tmpdict[mirnaid] = value
        del ortho_dict
        ortho_dict = tmpdict

    # count species and skip species with few orthologs before aligning
    if auto_skip:
        spec_counter = Counter(spec_list)
        max_spec_count = spec_counter.most_common(1)[0][1]
        count_cutoff = max_spec_count * auto_skip
        print(f'# Only species with at least {round(count_cutoff)} identified orthologs are used for tree calculation')
        for species, count in dict(spec_counter).items():
            if count < count_cutoff:
                spec_to_skip.append(species)
        print('# Skipping these species:')
        for spsk in spec_to_skip:
            print(spsk)

    # if no species list is given to include, all species are part of specinclude
    if not spec_include:
        spec_include = list(ortho_dict.keys())

    # make alignments
    print('# Starting alignments')
    align_out = f'{outdir}/alignments'
    if not os.path.isdir(align_out):
        os.mkdir(align_out)
    for mirna in ortho_dict:
        with tempfile.NamedTemporaryFile(mode='w+') as fp:
            for spec, seq in ortho_dict[mirna].items():
                if spec_to_skip:
                    if (
                            spec in spec_to_skip
                            and spec not in spec_include
                    ):
                        continue
                fp.write(f'>{spec}\n{seq}\n')
            fp.seek(0)
            aln_cmd = f'muscle -align {fp.name} -output {align_out}/{mirna}.aln'
            res = sp.run(aln_cmd, shell=True, capture_output=True)
            if res.returncode != 0:
                print(res.stderr.decode('utf8'))
                sys.exit()

    sys.stdout.flush()
    superm_path = make_supermatrix(outdir)
    print('# Starting tree calculation')
    tree_out = f'{outdir}/species_tree'
    if not os.path.isdir(tree_out):
        os.mkdir(tree_out)
    tree_cmd = iqtree_cmd.format(superm_path, tree_out)
    res = sp.run(tree_cmd, shell=True, capture_output=True)
    print(res.stdout.decode('utf-8'))
    if res.returncode != 0:
        print(res.stderr.decode('utf-8'))

    # write tree that can be used as optional phyloprofile input
    treepath = f'{tree_out}/species_tree.contree'
    pptree = f'{outdir}/PhyloProfile_tree.contree'
    with open(treepath, 'r') as th:
        treestr = th.read().strip()
        for name in name_2_taxid:
            ppid = f'ncbi{name_2_taxid[name]}'
            treestr = treestr.replace(name, ppid)
    with open(pptree, 'w') as otfh:
        otfh.write(treestr)


if __name__ == "__main__":
    main()



import os
import pyfaidx
try:
    from ncOrtho.coreset.coreset_utils import parse_annotation
except ModuleNotFoundError:
    from ncOrtho.coreset.coreset_utils import parse_annotation


def vprint(s, verbose):
    if verbose:
        print(s, flush=True)


def read_genome(path, coretax, outdir):
    core_gen_dir = f'{outdir}/core_genomes'
    if not os.path.isdir(core_gen_dir):
        os.mkdir(core_gen_dir)
    slink = f'{core_gen_dir}/{coretax}'
    try:
        os.symlink(path, slink)
    except FileExistsError:
        pass
    g = pyfaidx.Fasta(slink)
    return g


def synteny_check(left: list, right: list, orthodict: dict, mgi: int, v):
    """
    TODO: Will extract all syntenic regions using all available anchors. Should only return smallest region per co-ortholog
    Parameters
    ----------
    left
    right
    orthodict
    mgi
    v

    Returns
    -------

    """
    seqcol = []
    for left_ortho in left:
        if left_ortho not in orthodict:
            continue
        left_chromosome, left_position = orthodict[left_ortho]
        for right_ortho in right:
            if right_ortho not in orthodict:
                continue
            right_chromosome, right_position = orthodict[right_ortho]
            distance = abs(left_position - right_position)
            if left_chromosome == right_chromosome and distance <= mgi:
                leftstart, leftend, leftstrand = orthodict[left_chromosome][left_position][1:4]
                rightstart, rightend, rightstrand = orthodict[left_chromosome][right_position][1:4]
                #vprint('Synteny fulfilled', v)
                if left_position < right_position:
                    seqcol.append((left_chromosome, leftend, rightstart))
                else:
                    seqcol.append((left_chromosome, rightend, leftstart))
    return seqcol


def analyze_synteny(core_d, mirna_pos, out, idtype, mgi, v):
    """

    Parameters
    ----------
    core_d
    mirna_pos
    out
    idtype
    mgi
    v

    Returns:    {mirid: ['>coretaxon1\n', 'seq1\n', 'coretaxon2\n', 'seq2\n']}
    -------

    """
    synteny_region_collector = {}
    for taxon in core_d:
        print(f'# {taxon}', flush=True)
        vprint(f'# Parsing annotation file for {taxon}', v)

        core_anno_dict = parse_annotation(core_d[taxon]['annotation'], idtype)

        vprint('# Loading genome file', v)
        fasta_path = core_d[taxon]['genome']
        genome = read_genome(fasta_path, taxon, out)
        vprint('# Done', v)

        for mirid, positiontuple in mirna_pos.items():
            synteny_fulfilled = False
            if mirid not in synteny_region_collector:
                synteny_region_collector[mirid] = []
            style, core_ortholog_collection_all_taxa = positiontuple
            if taxon not in core_ortholog_collection_all_taxa:
                vprint(f'No core orthologs found for {mirid} in {taxon}', v)
                continue
            core_ortholog_collection = core_ortholog_collection_all_taxa[taxon]

            leftorthos, rightorthos = core_ortholog_collection
            #print(style)
            if leftorthos == rightorthos:  # for genes inside a gene for which an ortholog was found
                #if style in ['inside', 'opposite']:
                for ortholog in leftorthos:
                    if ortholog in core_anno_dict:
                        ortholog_chromosome, ortholog_position = core_anno_dict[ortholog]
                        orthostart, orthoend, orthostrand = core_anno_dict[ortholog_chromosome][ortholog_position][1:]
                        seq = genome[ortholog_chromosome][orthostart-1:orthoend].seq

                        synteny_fulfilled = True
                        synteny_region_collector[mirid].append(f'>{taxon}_0\n')
                        synteny_region_collector[mirid].append(f'{seq}\n')
                        # ToDo: check if opposite needs to be reverse
                        # if style == 'inside':
                        #     seq = genome[ortholog_chromosome][orthostart-orthoend].seq
                        # else:
                        #     seq = genome[ortholog_chromosome][orthostart - orthoend].reverse.complement.seq
            else:
                leftorthos, rightorthos = core_ortholog_collection
                synregs = synteny_check(leftorthos, rightorthos, core_anno_dict, mgi, v)
                if not synregs:
                    continue
                for count, syntenyregion in enumerate(synregs):
                    synchrom, synstart, synend = syntenyregion
                    seq = genome[synchrom][synstart-1:synend].seq

                    synteny_fulfilled = True
                    synteny_region_collector[mirid].append(f'>{taxon}_{count}\n')
                    synteny_region_collector[mirid].append(f'{seq}\n')
            if synteny_fulfilled:
                vprint(f'Synteny fulfilled for {mirid}', v)
            else:
                vprint(f'No syntenic region found for {mirid}', v)
    return synteny_region_collector
